/****************************** comma.c
 look for a specific scan-code
 and exchange the key-code


 Walter Unterberger, 1992

 program is derived from examples of the excellent book:
 "Extending Turbo C Professional" by Al Stevens

 compiled with Borland C++ 3.0
 with Test Stack overflow turned off
*************************/

#include <stdio.h>
#include <conio.h>
#include <dos.h>
#include <ctype.h>
#include "interupt.h"

#define	KOMMA	','
#define DOT	'.'
#define DEL_SCAN	83
#define FALSE	0
#define TRUE	1

static void interrupt (*oldkb)(void);
static int far *nextoff = MK_FP(0x40,0x1a);
static int far *nexton  = MK_FP(0x40,0x1c);
static unsigned mcbseg;
char signature[] = "COMMA";

/* reduce heaplength and stacklength to make a smaller program! */
extern unsigned _heaplen = 512;
extern unsigned _stklen = 512;

/* --- stuff a key value into the BIOS read-ahead buffer --- */
void stuffkey(int k)
{
	if (*nexton+2 == *nextoff ||
		(*nexton == 0x3c && *nextoff == 0x1e))
			return;
	poke(0x40, *nexton, k);
	*nexton += 2;
	if (*nexton == 0x3e)
		*nexton = 0x1e;
}

/* -------- clear the BIOS read-ahead buffer ---------- */
void clear_biosbuff(void)
{
	*nextoff = *nexton;
}

/* ----- ISR to translate non-keys to keys
*/
static void interrupt newkb()
{
	static int scancode, shiftkeys, kbcntl;

	scancode = inportb(KBDATA);
	shiftkeys = peekb(0x40, 0x17);

	if ( (shiftkeys & 0x20) && (scancode == DEL_SCAN))
	      {	stuffkey(DOT);
		/* now reset keyboard	*/
		kbcntl = inportb(KBCONTROL);
		outportb(KBCONTROL, kbcntl | 0x80);
		outportb(KBCONTROL, kbcntl);
		outportb(0x20, 0x20);
	      }
	else
	      (*oldkb)();
}


/* ---- test if the program is already resident
*/
static int resident(char *signature)
{
	char *sg;
	unsigned df;
	unsigned blkseg, mcbs = mcbseg;

	df = _DS - _psp;
	/* ---- walk thru mcb chain & search for TSR ----- */
	while (peekb(mcbs, 0) == 0x4d)  {
		blkseg = peek(mcbs, 1);
		if (peek(blkseg, 0) == 0x20cd)  {
			/* --- this is a psp --- */
			if (blkseg == _psp)
				break;           /* if the transient copy */
			for (sg = signature; *sg; sg++)
				if (*sg != peekb(blkseg+df,(unsigned)sg))
					break;
			if (*sg == '\0')
				/* ---- TSR is already resident --- */
				return TRUE;
		}
		mcbs += peek(mcbs, 3) + 1;
	}
	return FALSE;
}


/* ------- now the main TSR routine
*/
void main(void)
{
    int c;
    unsigned es, bx;

    /* first get the seg address of first DOS MCB  */
    _AH = 0x52;
    geninterrupt(DOS);
    es = _ES;
    bx = _BX;
    mcbseg = peek(es, bx-2);

    /* look if already loaded	*/
    if (resident(signature))
       { puts("COMMA already loaded - no action taken");
	 exit(0);
       }

    /* now exchange keyboard ISR	*/
    oldkb = getvect(KYBRD);
    clear_biosbuff();
    setvect(KYBRD, newkb);

    puts("Numeric Keypad Dot driver loaded!");
    keep(0, (_SS + (_SP/16+10) - _psp));

}
